from typing import Any, Dict, Optional

from broadcaster import Event as BroadcasterEvent


class Event(BroadcasterEvent):
    def __init__(
        self,
        channel: str,
        message: Any,
        context: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(channel, message)

        if context is None:
            context = {}

        self.context = context



"""
asyncapi
"""

__version__ = '0.12.0'
from .api import AsyncApi, OperationsTypeHint
from .builder import (
    build_api,
    build_api_auto_spec,
    build_channel_operations,
    build_spec,
    dict_from_ref,
    fill_refs,
)
from .events import Event
from .events.handler import EventsHandler
from .exceptions import (
    ChannelOperationNotFoundError,
    GCloudPubSubConsumerDisconnectError,
    GCloudPubSubPublishTimeoutError,
    InvalidChannelError,
    OperationIdNotFoundError,
    ReferenceNotFoundError,
    UrlOrModuleRequiredError,
)
from .specification_v2_0_0 import (
    AutoSpec,
    Channel,
    Components,
    Info,
    Message,
    Operation,
    ProtocolType,
    Server,
    Specification,
)
from .subscriber import run as run_subscriber


from .docs import run as run_docs  # isort: skip


__all__ = [
    'Channel',
    'Components',
    'Info',
    'Message',
    'ProtocolType',
    'Server',
    'Specification',
    'Operation',
    'AutoSpec',
    'AsyncApi',
    'OperationsTypeHint',
    'build_api',
    'build_api_auto_spec',
    'build_spec',
    'build_channel_operations',
    'dict_from_ref',
    'fill_refs',
    'ChannelOperationNotFoundError',
    'InvalidChannelError',
    'OperationIdNotFoundError',
    'ReferenceNotFoundError',
    'UrlOrModuleRequiredError',
    'run_subscriber',
    'run_docs',
    'EventsHandler',
    'Event',
    'GCloudPubSubPublishTimeoutError',
    'GCloudPubSubConsumerDisconnectError',
]



import dataclasses
import importlib
import json
from enum import Enum
from http import HTTPStatus
from typing import Any, Dict, Iterable, List, Optional

import typer
import uvicorn
import yaml
from apidaora import (
    ContentType,
    Response,
    RoutedControllerTypeHint,
    appdaora,
    route,
)

from asyncapi import Operation, Specification
from asyncapi.builder import build_spec_from_path
from asyncapi.schema import type_as_jsonschema
from asyncapi.specification_v2_0_0 import as_camel_case


def main(
    api_module: str = typer.Option('', envvar='ASYNCAPI_MODULE'),
    host: str = typer.Option('0.0.0.0', envvar='ASYNCAPI_HOST'),
    port: int = typer.Option(5000, envvar='ASYNCAPI_PORT'),
    path: Optional[str] = typer.Option(None, envvar='ASYNCAPI_PATH'),
) -> None:
    if path:
        spec = build_spec_from_path(path)
    else:
        spec = getattr(importlib.import_module(api_module), 'spec')

    start(spec, host, port)


def start(spec: Specification, host: str, port: int) -> None:
    app = appdaora(
        build_yaml_spec_controllers(spec) + [build_json_spec_controller(spec)]
    )
    uvicorn.run(app, host=host, port=port)


def build_yaml_spec_controllers(
    spec: Specification,
) -> List[RoutedControllerTypeHint]:
    def controller() -> Response:
        return Response(
            status=HTTPStatus.OK,
            content_type=ContentType.APPLICATION_YAML,
            body=yaml.dump(spec_asjson(spec)),
            headers=(),
        )

    @route.get('/asyncapi.yaml')
    def controller_yaml() -> Response:
        return controller()

    @route.get('/asyncapi.yml')
    def controller_yml() -> Response:
        return controller()

    return [controller_yaml, controller_yml]


def build_json_spec_controller(
    spec: Specification,
) -> RoutedControllerTypeHint:
    @route.get('/asyncapi.json')
    def controller() -> Response:
        return Response(
            status=HTTPStatus.OK,
            content_type=ContentType.APPLICATION_JSON,
            body=json.dumps(spec_asjson(spec), indent=2).encode(),
            headers=(),
        )

    return controller


def spec_asjson(spec: Specification) -> Dict[str, Any]:
    json_spec: Dict[str, Any] = _spec_asjson(spec)
    spec_messages_dict = (
        spec.components.messages
        if spec.components and spec.components.messages
        else {}
    )
    json_spec_messages = {}

    for message_name, message_type in spec_messages_dict.items():
        if message_type.payload:
            json_spec['components']['messages'][message_name][
                'payload'
            ] = type_as_jsonschema(message_type.payload)

            if message_type.name:
                json_spec_messages[message_type.name] = message_name

    for server in json_spec['servers'].values():
        server.pop('name', None)

    for channel_dict, channel in zip(
        json_spec['channels'].values(), spec.channels.values()
    ):
        channel_dict.pop('name', None)

        if 'subscribe' in channel_dict:
            set_operation_message(
                channel_dict['subscribe'],
                channel.subscribe,
                json_spec_messages,
            )

        if 'publish' in channel_dict:
            set_operation_message(
                channel_dict['publish'], channel.publish, json_spec_messages
            )

    return json_spec


def set_operation_message(
    operation_dict: Dict[str, Any],
    operation: Optional[Operation],
    json_spec_messages: Dict[str, str],
) -> None:
    operation_dict['message'].pop('contentType', None)

    if (
        operation
        and operation.message
        and operation.message.name
        and operation.message.name in json_spec_messages
    ):
        operation_dict['message'] = {
            '$ref': (
                '#/components/messages/'
                f'{json_spec_messages[operation.message.name]}'
            )
        }

    elif operation and operation.message and operation.message.payload:
        operation_dict['message']['payload'] = type_as_jsonschema(
            operation.message.payload
        )


def _spec_asjson(generic_value: Any) -> Any:
    json_value: Any

    if dataclasses.is_dataclass(generic_value):
        json_value = {}

        for field in dataclasses.fields(generic_value):
            field_value = _spec_asjson(
                getattr(generic_value, field.name, None)
            )

            if field_value is not None and field_value != '':
                json_value[as_camel_case(field.name)] = field_value

    elif isinstance(generic_value, dict):
        json_value = {k: _spec_asjson(v) for k, v in generic_value.items()}

    elif not isinstance(generic_value, str) and isinstance(
        generic_value, Iterable
    ):
        json_value = [_spec_asjson(v) for v in generic_value]

    elif isinstance(generic_value, Enum):
        json_value = generic_value.value

    else:
        json_value = generic_value

    return json_value


def run() -> None:
    typer.run(main)


if __name__ == '__main__':
    run()



import asyncio
import functools
import secrets
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as FutureTimeoutError
from logging import Logger, getLogger
from typing import TYPE_CHECKING, Any, Dict, Tuple
from urllib.parse import urlparse

from broadcaster._backends.base import BroadcastBackend
from google.cloud import pubsub_v1
from google.cloud.pubsub_v1.types import PullResponse, ReceivedMessage

from asyncapi import (
    GCloudPubSubConsumerDisconnectError,
    GCloudPubSubPublishTimeoutError,
)

from .. import Event


if TYPE_CHECKING:
    AsyncFutureHint = asyncio.Future[PullResponse]
else:
    AsyncFutureHint = asyncio.Future


class GCloudPubSubBackend(BroadcastBackend):
    def __init__(
        self,
        url: str,
        bindings: Dict[str, str] = {},
        logger: Logger = getLogger(__name__),
    ):
        url_parsed = urlparse(url, scheme='gcloud-pubsub')
        self._project = url_parsed.netloc
        self._consumer_channels: Dict[str, str] = {}
        self._producer_channels: Dict[str, str] = {}
        self._channel_index = 0
        self._logger = logger
        self._set_consumer_config(bindings)
        self._disconnected = True
        self._executor = ThreadPoolExecutor(self._consumer_max_workers)

    async def connect(self) -> None:
        self._producer = pubsub_v1.PublisherClient()
        self._consumer = pubsub_v1.SubscriberClient()
        self._disconnected = False

    async def disconnect(self) -> None:
        self._disconnected = True
        self._producer.stop()
        self._consumer.close()
        del self._producer
        del self._consumer

    async def subscribe(self, channel: str) -> None:
        pubsub_channel = self._consumer.subscription_path(
            self._project, channel
        )
        self._consumer_channels[channel] = pubsub_channel

    async def unsubscribe(self, channel: str) -> None:
        self._consumer_channels.pop(channel)

    async def publish(
        self, channel: str, message: Any, retries_counter: int = 1
    ) -> None:
        producer_channel = self._producer_channels.get(channel)

        if not producer_channel:
            producer_channel = self._producer.topic_path(
                self._project, channel
            )
            self._producer_channels[channel] = producer_channel

        future = self._producer.publish(producer_channel, message.encode())

        try:
            future.result(timeout=self._publish_timeout)
        except FutureTimeoutError:
            if retries_counter >= self._publish_retries:
                raise GCloudPubSubPublishTimeoutError(
                    f'publish timeout; channel={channel}; '
                    f'message={message[:100]}...'
                )
            else:
                await self.publish(channel, message, retries_counter + 1)

    async def next_published(self) -> Event:
        (
            received_message,
            channel_id,
            pubsub_channel,
        ) = await self._pull_message_from_consumer()
        event = Event(channel_id, received_message.message.data.decode())

        if self._consumer_ack_messages:
            await self.wait_ack(received_message, pubsub_channel)
        else:
            event.context['ack_func'] = functools.partial(
                self.wait_ack, received_message, pubsub_channel,
            )

        return event

    async def _pull_message_from_consumer(
        self,
    ) -> Tuple[ReceivedMessage, str, str]:
        channel_index = (
            secrets.choice(range(len(self._consumer_channels)))
            if self._consumer_channels
            else 0
        )

        while not self._disconnected:
            channels = list(self._consumer_channels.items())

            if not len(channels):
                await asyncio.sleep(self._consumer_wait_time)
                continue

            if channel_index >= len(channels):
                channel_index = 0

            channel_id, pubsub_channel = channels[channel_index]
            pull_message_future: AsyncFutureHint
            pull_message_future = asyncio.get_running_loop().run_in_executor(  # type: ignore
                self._executor,
                functools.partial(
                    self._consumer.pull,
                    pubsub_channel,
                    max_messages=1,
                    return_immediately=True,
                ),
            )

            try:
                response = await asyncio.wait_for(
                    pull_message_future, self._consumer_pull_message_timeout
                )

            except asyncio.TimeoutError:
                channel_index += 1
                await asyncio.sleep(self._consumer_wait_time)
                continue

            else:
                if not response.received_messages:
                    channel_index += 1
                    await asyncio.sleep(self._consumer_wait_time)
                    continue

                await asyncio.sleep(self._pull_message_wait_time)
                return (
                    response.received_messages[0],
                    channel_id,
                    pubsub_channel,
                )

        raise GCloudPubSubConsumerDisconnectError()

    async def wait_ack(
        self,
        message: ReceivedMessage,
        pubsub_channel: str,
        retries_counter: int = 1,
    ) -> None:
        future = asyncio.get_running_loop().run_in_executor(
            self._executor,
            functools.partial(
                self._consumer.acknowledge, pubsub_channel, [message.ack_id],
            ),
        )

        try:
            await asyncio.wait_for(future, timeout=self._consumer_ack_timeout)
        except asyncio.TimeoutError:
            if retries_counter >= self._consumer_ack_retries:
                self._logger.warning(
                    f'ack timeout {self._consumer_ack_timeout}; '
                    f'message={message.message.data.decode()[:100]}...'
                )
            else:
                await self.wait_ack(
                    message, pubsub_channel, retries_counter + 1
                )
        except asyncio.CancelledError:
            self._logger.warning(
                'ack cancelled; '
                f'message={message.message.data.decode()[:100]}...'
            )

    def _set_consumer_config(self, bindings: Dict[str, str]) -> None:
        consumer_wait_time = 1.0
        consumer_ack_messages = False
        consumer_ack_timeout = 1.0
        consumer_ack_retries = 3
        consumer_pull_message_timeout = 1.0
        consumer_max_workers = 10
        publish_timeout = 5.0
        publish_retries = 3
        pull_message_wait_time = 0.1

        for config_name, config_value in bindings.items():
            if config_name == 'consumer_wait_time':
                consumer_wait_time = float(config_value)

            elif config_name == 'consumer_ack_messages':
                consumer_ack_messages = config_value in (
                    '1',
                    'true',
                    't',
                    'True',
                    'y',
                    'yes',
                )

            elif config_name == 'consumer_ack_timeout':
                consumer_ack_timeout = float(config_value)

            elif config_name == 'consumer_ack_retries':
                consumer_ack_retries = int(config_value)

            elif config_name == 'consumer_max_workers':
                consumer_max_workers = int(config_value)

            elif config_name == 'consumer_pull_message_timeout':
                consumer_pull_message_timeout = float(config_value)

            elif config_name == 'publish_timeout':
                publish_timeout = float(config_value)

            elif config_name == 'publish_retries':
                publish_retries = int(config_value)

            elif config_name == 'pull_message_wait_time':
                pull_message_wait_time = float(config_value)

        self._consumer_wait_time = consumer_wait_time
        self._consumer_ack_messages = consumer_ack_messages
        self._consumer_ack_timeout = consumer_ack_timeout
        self._consumer_ack_retries = consumer_ack_retries
        self._consumer_pull_message_timeout = consumer_pull_message_timeout
        self._consumer_max_workers = consumer_max_workers
        self._publish_timeout = publish_timeout
        self._publish_retries = publish_retries
        self._pull_message_wait_time = pull_message_wait_time



from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Type, get_type_hints


DEFAULT_CONTENT_TYPE = 'application/json'
ASYNCAPI_VERSION = '2.0.0'


@dataclass
class Specification:
    info: 'Info'
    channels: Dict[str, 'Channel']
    servers: Optional[Dict[str, 'Server']] = None
    components: Optional['Components'] = None
    tags: Optional[List['Tag']] = None
    external_docs: Optional['ExternalDocumentation'] = None
    default_content_type: str = DEFAULT_CONTENT_TYPE
    asyncapi: str = ASYNCAPI_VERSION


@dataclass(init=False)
class AutoSpec(Specification):
    def __init__(
        self,
        title: str,
        description: str = '',
        version: str = '1.0.0',
        **servers: str,
    ) -> None:
        self.info = Info(title, version, description)
        self.servers = {}
        self.channels = {}

        for server_name, url in servers.items():
            protocol, uri = url.split('://')
            self.servers[server_name] = Server(
                url=uri, protocol=ProtocolType(protocol), name=server_name,
            )

    def subscribe(
        self,
        subscbriber_: Optional[Callable[..., Any]] = None,
        *,
        channel_name: str,
        channel_description: str = '',
        message_name: str = '',
        message_title: str = '',
        message_summary: str = '',
    ) -> Callable[..., Callable[..., Any]]:
        if not message_name:
            message_name = channel_name

        message_name = message_name.replace('/', '_')
        message_name = message_name.replace('#', '_')
        message_name = message_name.replace(' ', '_')
        message_name = as_camel_case(message_name)
        message_component_name = message_name[0].upper() + message_name[1:]

        def decorator(subscbriber: Callable[..., Any]) -> Callable[..., Any]:
            message_type = get_type_hints(subscbriber).get('message')
            message = Message(
                name=message_name,
                title=message_title,
                summary=message_summary,
                payload=message_type,
            )
            self.channels[channel_name] = subscbriber.__channel__ = Channel(  # type: ignore
                description=channel_description,
                subscribe=Operation(
                    operation_id=subscbriber.__name__, message=message,
                ),
                publish=Operation(message=message),
                name=channel_name,
            )
            if self.components and self.components.messages:
                self.components.messages[message_component_name] = message

            elif self.components:
                self.components.messages = {message_component_name: message}

            else:
                self.components = Components(
                    messages={message_component_name: message}
                )

            return subscbriber

        if subscbriber_:
            return decorator(subscbriber_)
        else:
            return decorator


@dataclass
class Contact:
    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@dataclass
class License:
    name: str
    url: Optional[str] = None


@dataclass
class Info:
    title: str
    version: str
    description: Optional[str] = None
    terms_of_service: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None


class ProtocolType(Enum):
    KAFKA = 'kafka'
    REDIS = 'redis'
    POSTGRES = 'postgres'
    GCLOUD_PUBSUB = 'gcloud-pubsub'


@dataclass
class ServerVariable:
    enum: Optional[List[str]] = None
    default: Optional[str] = None
    description: Optional[str] = None
    examples: Optional[List[str]] = None


@dataclass
class Server:
    url: str
    protocol: ProtocolType
    protocol_version: Optional[str] = None
    description: Optional[str] = None
    variables: Optional[Dict[str, ServerVariable]] = None
    security: Optional[Dict[str, List[str]]] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None
    name: Optional[str] = None


@dataclass
class CorrelationId:
    location: str
    description: Optional[str] = None


@dataclass
class ExternalDocumentation:
    url: str
    description: Optional[str] = None


@dataclass
class Tag:
    name: str
    description: Optional[str] = None
    external_docs: Optional[ExternalDocumentation] = None


@dataclass
class MessageTrait:
    headers: Optional[Dict[str, Any]] = None
    correlation_id: Optional[CorrelationId] = None
    schema_format: Optional[str] = None
    content_type: Optional[str] = None
    name: Optional[str] = None
    title: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[Tag]] = None
    external_docs: Optional[ExternalDocumentation] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None
    examples: Optional[List[Dict[str, Any]]] = None


@dataclass
class Message:
    headers: Optional[Dict[str, Any]] = None
    payload: Optional[Type[Any]] = None
    correlation_id: Optional[CorrelationId] = None
    schema_format: Optional[str] = None
    content_type: Optional[str] = None
    name: Optional[str] = None
    title: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[Tag]] = None
    external_docs: Optional[ExternalDocumentation] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None
    examples: Optional[List[Dict[str, Any]]] = None
    traits: Optional[List[MessageTrait]] = None


@dataclass
class OperationTrait:
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[Tag]] = None
    external_docs: Optional[ExternalDocumentation] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None


@dataclass
class Operation:
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[Tag]] = None
    external_docs: Optional[ExternalDocumentation] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None
    traits: Optional[List[OperationTrait]] = None
    message: Optional[Message] = None


@dataclass
class Parameter:
    description: Optional[str] = None
    schema: Optional[Type[Any]] = None
    location: Optional[str] = None


@dataclass
class Channel:
    description: Optional[str] = None
    publish: Optional[Operation] = None
    subscribe: Optional[Operation] = None
    parameters: Optional[Dict[str, Parameter]] = None
    bindings: Optional[Dict[ProtocolType, Any]] = None
    name: Optional[str] = None


@dataclass
class Components:
    messages: Optional[Dict[str, Message]] = None
    schemas: Optional[Dict[str, Any]] = None


def as_camel_case(snake_str: str) -> str:
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])



import asyncio
import dataclasses
import logging
from typing import Any, Callable, Dict, Optional, Tuple, Type

import orjson
from jsondaora import DeserializationError, asdataclass, dataclass_asjson

from .events.handler import EventsHandler
from .exceptions import (
    ChannelOperationNotFoundError,
    ChannelPublishNotFoundError,
    InvalidChannelError,
    InvalidMessageError,
    OperationIdNotFoundError,
)
from .specification_v2_0_0 import Operation, Specification


OperationsTypeHint = Dict[Tuple[str, str], Callable[..., Any]]


@dataclasses.dataclass
class AsyncApi:
    spec: Specification
    operations: OperationsTypeHint
    events_handler: EventsHandler
    republish_error_messages: bool = False
    republish_error_messages_channels: Optional[Dict[str, str]] = None
    logger: logging.Logger = logging.getLogger(__name__)

    async def publish_json(
        self, channel_id: str, message: Dict[str, Any]
    ) -> None:
        await self.events_handler.publish(
            channel=channel_id,
            message=self.parse_message(
                channel_id, self.payload(channel_id, **message)
            ).decode(),
        )

    async def publish(self, channel_id: str, message: Any) -> None:
        await self.events_handler.publish(
            channel=channel_id,
            message=self.parse_message(channel_id, message).decode(),
        )

    async def __aenter__(self) -> 'AsyncApi':
        await self.connect()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        await self.events_handler.connect()

    async def disconnect(self) -> None:
        await self.events_handler.disconnect()

    def payload(self, channel_id: str, **message: Any) -> Any:
        type_ = self.publish_payload_type(channel_id)
        return self.payload_type(type_, channel_id, **message)

    def subscriber_payload(self, channel_id: str, **message: Any) -> Any:
        type_ = self.subscribe_payload_type(channel_id)
        return self.payload_type(type_, channel_id, **message)

    def payload_type(
        self, type_: Type[Any], channel_id: str, **message: Any
    ) -> Any:
        if type_ and dataclasses.is_dataclass(type_):
            return asdataclass(message, type_)

        return message

    async def listen_all(self) -> None:
        tasks = []

        for channel_id, channel in self.spec.channels.items():
            if channel.subscribe:
                task = asyncio.create_task(self.listen(channel_id))
                task.add_done_callback(task_callback)
                tasks.append(task)

        await asyncio.gather(*tasks)

    async def listen(self, channel_id: str) -> None:
        if self.spec.channels is None:
            raise InvalidChannelError(channel_id)

        operation_id = self.subscribe_operation(channel_id).operation_id

        if operation_id is None:
            raise ChannelOperationNotFoundError(channel_id)

        try:
            operation_func = self.operations[(channel_id, operation_id)]
        except KeyError:
            raise OperationIdNotFoundError(channel_id, operation_id)

        async with self.events_handler.subscribe(
            channel=channel_id
        ) as subscriber:
            async for event in subscriber:
                try:
                    json_message = orjson.loads(event.message)
                    payload = self.subscriber_payload(
                        channel_id, **json_message
                    )

                    coro = operation_func(
                        payload, **getattr(event, 'context', {})
                    )

                    while asyncio.iscoroutine(coro):
                        coro = await coro

                except (orjson.JSONDecodeError, DeserializationError):
                    self.logger.exception(f"message={event.message[:100]}")

                except Exception:
                    self.logger.exception(f"message={event.message[:100]}")

                    if self.republish_error_messages:
                        republish_channel = (
                            self.republish_error_messages_channels.get(
                                channel_id, channel_id
                            )
                            if self.republish_error_messages_channels
                            is not None
                            else channel_id
                        )

                        try:
                            await self.publish(republish_channel, payload)
                        except UnboundLocalError:
                            await self.publish_json(
                                republish_channel, json_message
                            )
                        except Exception:
                            self.logger.exception(
                                f"message={event.message[:100]}"
                            )

    def publish_operation(self, channel_id: str) -> Operation:
        return self.operation('publish', channel_id)

    def subscribe_operation(self, channel_id: str) -> Operation:
        return self.operation('subscribe', channel_id)

    def operation(self, op_name: str, channel_id: str) -> Operation:
        operation: Operation

        try:
            operation = getattr(self.spec.channels[channel_id], op_name)

            if operation is None:
                raise ChannelPublishNotFoundError(channel_id)

        except KeyError:
            raise InvalidChannelError(channel_id)

        else:
            return operation

    def parse_message(self, channel_id: str, message: Any) -> Any:
        type_ = self.publish_payload_type(channel_id)

        if type_:
            if not isinstance(message, type_):
                raise InvalidMessageError(message, type_)

            return dataclass_asjson(message)

        return message

    def publish_payload_type(self, channel_id: str) -> Any:
        operation = self.publish_operation(channel_id)

        if operation.message is None:
            return None

        return operation.message.payload

    def subscribe_payload_type(self, channel_id: str) -> Any:
        operation = self.subscribe_operation(channel_id)

        if operation.message is None:
            return None

        return operation.message.payload


def task_callback(future: Any) -> None:
    future.result()



"""
asyncapi
"""
import importlib
import io
from collections import defaultdict, deque
from typing import Any, DefaultDict, Dict, Optional

import requests
import yaml
from jsondaora import jsonschema_asdataclass

from .api import AsyncApi, OperationsTypeHint
from .events.handler import EventsHandler
from .exceptions import (
    EmptyServersError,
    InvalidAsyncApiVersionError,
    InvalidChannelsSubscribersError,
    InvalidContentTypeError,
    InvalidServerBindingError,
    InvalidServerBindingProtocolError,
    ReferenceNotFoundError,
    ServerNotFoundError,
)
from .specification_v2_0_0 import (
    ASYNCAPI_VERSION,
    DEFAULT_CONTENT_TYPE,
    Channel,
    Components,
    Info,
    Message,
    Operation,
    ProtocolType,
    Server,
    Specification,
)


def build_api(
    path: str,
    server: Optional[str] = None,
    module_name: str = '',
    republish_errors: Optional[bool] = None,
    server_bindings: Optional[str] = None,
    channels_subscribes: Optional[str] = None,
    republish_errors_channels: Optional[str] = None,
) -> AsyncApi:
    spec = build_spec_from_path(path)
    set_api_spec_server_bindings(spec, server_bindings)
    set_api_spec_channels_subscribes(spec, channels_subscribes)
    return build_api_from_spec(
        spec, module_name, server, republish_errors, republish_errors_channels
    )


def build_api_auto_spec(
    module_name: str,
    server: Optional[str] = None,
    republish_errors: Optional[bool] = None,
    server_bindings: Optional[str] = None,
    channels_subscribes: Optional[str] = None,
    republish_errors_channels: Optional[str] = None,
) -> AsyncApi:
    spec = getattr(importlib.import_module(module_name), 'spec')
    set_api_spec_server_bindings(spec, server_bindings)
    set_api_spec_channels_subscribes(spec, channels_subscribes)
    return build_api_from_spec(
        spec, module_name, server, republish_errors, republish_errors_channels
    )


def build_spec_from_path(path: str) -> Specification:
    return build_spec(load_spec_dict(path))


def set_api_spec_server_bindings(
    spec: Specification, server_bindings: Optional[str]
) -> None:
    if server_bindings and spec.servers:
        binding_str_list = server_bindings.split(',')
        bindings_spec: DefaultDict[str, Dict[str, str]] = defaultdict(dict)

        for binding_str in binding_str_list:
            try:
                protocol_str, protocol_bind_str = binding_str.split(':')
            except ValueError:
                raise InvalidServerBindingError(binding_str)

            try:
                protocol_bind_list = protocol_bind_str.split(';')
            except ValueError:
                raise InvalidServerBindingError(binding_str)

            for protocol_bind_str in protocol_bind_list:
                try:
                    binding_name, binding_value = protocol_bind_str.split('=')
                except ValueError:
                    raise InvalidServerBindingError(protocol_bind_str)

                bindings_spec[protocol_str][binding_name] = binding_value

        for server_name, server in spec.servers.items():
            new_server_bindings = build_server_bindings(
                server_name, server.protocol, bindings_spec
            )

            if new_server_bindings:
                for protocol, binding in new_server_bindings.items():
                    already_server_bindings = server.bindings

                    if already_server_bindings:
                        if protocol in already_server_bindings:
                            already_server_bindings[protocol].update(binding)
                        else:
                            already_server_bindings[protocol] = binding

                    else:
                        server.bindings = {protocol: binding}


def set_api_spec_channels_subscribes(
    spec: Specification, channels_subscribes: Optional[str]
) -> None:
    if channels_subscribes:
        channels_subscribes_list = channels_subscribes.split(',')
        channels_subscribes_dict = {}
        new_channels = {}

        for channels_subscribes_str in channels_subscribes_list:
            try:
                (
                    publish_channel,
                    subscribe_operation_id,
                ) = channels_subscribes_str.split(':')
            except ValueError:
                InvalidChannelsSubscribersError(channels_subscribes_str)

            if '=' in subscribe_operation_id:
                try:
                    (
                        subscribe_channel,
                        subscribe_operation_id,
                    ) = subscribe_operation_id.split('=')
                except ValueError:
                    InvalidChannelsSubscribersError(channels_subscribes_str)
            else:
                subscribe_channel = publish_channel

            channels_subscribes_dict[publish_channel] = (
                subscribe_channel,
                subscribe_operation_id,
            )

        for channel_name, channel in spec.channels.items():
            if (
                channel.subscribe is None
                and channel.publish
                and channel_name in channels_subscribes_dict
            ):
                (
                    subscribe_channel_name,
                    operation_id,
                ) = channels_subscribes_dict[channel_name]
                subscribe = Operation(
                    operation_id=operation_id,
                    summary=channel.publish.summary,
                    description=channel.publish.description,
                    tags=channel.publish.tags,
                    external_docs=channel.publish.external_docs,
                    bindings=channel.publish.bindings,
                    traits=channel.publish.traits,
                    message=channel.publish.message,
                )

                if subscribe_channel_name == channel_name:
                    channel.subscribe = subscribe
                else:
                    new_channels[subscribe_channel_name] = Channel(
                        description=channel.description,
                        subscribe=subscribe,
                        parameters=channel.parameters,
                        bindings=channel.bindings,
                        name=subscribe_channel_name,
                    )

        spec.channels.update(new_channels)


def build_api_from_spec(
    spec: Specification,
    module_name: str,
    server_name: Optional[str],
    republish_errors: Optional[bool],
    republish_errors_channels: Optional[str],
) -> AsyncApi:
    if spec.servers is None or not spec.servers:
        raise EmptyServersError()

    if server_name is None:
        server_name = tuple(spec.servers.keys())[-1]

    try:
        server = spec.servers[server_name]
    except KeyError:
        ServerNotFoundError(server_name)

    if module_name:
        try:
            logger = getattr(importlib.import_module(module_name), 'logger')
        except (ImportError, AttributeError):
            logger = AsyncApi.logger
    else:
        logger = AsyncApi.logger

    operations = build_channel_operations(spec, module_name)
    events_handler = EventsHandler(
        url=f'{server.protocol.value}://{server.url}',
        bindings=server.bindings.get(server.protocol, {})
        if server.bindings
        else {},
    )

    republish_errors_channels_dict: Optional[Dict[str, str]]
    if republish_errors_channels is not None:
        republish_errors_channels_dict = {
            channel_str.split('=')[0]: channel_str.split('=')[1]
            for channel_str in republish_errors_channels.split(';')
        }
    else:
        republish_errors_channels_dict = None

    if republish_errors is not None:
        kwargs = dict(republish_error_messages=republish_errors)
    else:
        kwargs = {}

    return AsyncApi(
        spec,
        operations,
        events_handler,
        republish_error_messages_channels=republish_errors_channels_dict,
        logger=logger,
        **kwargs,
    )


def build_channel_operations(
    spec: Specification, module_name: str
) -> OperationsTypeHint:
    operations = {}

    if module_name:
        for channel_name, channel in spec.channels.items():
            if channel.subscribe and channel.subscribe.operation_id:
                operation = getattr(
                    importlib.import_module(module_name),
                    channel.subscribe.operation_id,
                    None,
                )

                if operation:
                    operation_key = (
                        channel_name,
                        channel.subscribe.operation_id,
                    )
                    operations[operation_key] = operation

    return operations


def load_spec_dict(path: str) -> Dict[str, Any]:
    spec: Dict[str, Any]

    if path.startswith('http'):
        request = requests.get(path)
        request.raise_for_status()

        if path.endswith('.json'):
            spec = request.json()
        else:
            spec = yaml.safe_load(io.BytesIO(request.content))

    else:
        spec = yaml.safe_load(open(path))

    return spec


def build_spec(spec: Dict[str, Any]) -> Specification:
    fill_refs(spec)
    validate_content_type(spec.get('defaultContentType', DEFAULT_CONTENT_TYPE))
    validate_asyncapi_version(spec.get('asyncapi', ASYNCAPI_VERSION))

    return Specification(
        info=Info(**spec['info']),
        servers={
            server_name: build_server(server_name, server_spec)
            for server_name, server_spec in spec['servers'].items()
        }
        if 'servers' in spec
        else None,
        channels=build_channels(spec),
        components=build_components(spec.get('components')),
    )


def build_server(server_name: str, server_spec: Dict[str, Any]) -> Server:
    protocol = ProtocolType(server_spec.pop('protocol'))
    bindings = build_server_bindings(
        server_name, protocol, server_spec.pop('bindings', None)
    )
    return Server(
        name=server_name, protocol=protocol, bindings=bindings, **server_spec,
    )


def build_server_bindings(
    server_name: str,
    server_protocol: ProtocolType,
    bindings_spec: Optional[Dict[str, Dict[str, str]]],
) -> Optional[Dict[ProtocolType, Dict[str, str]]]:
    if bindings_spec:
        server_bindings = {}

        for protocol_str, bindings in bindings_spec.items():
            try:
                binding_protocol = ProtocolType(protocol_str)
            except ValueError:
                raise InvalidServerBindingProtocolError(
                    server_name, protocol_str
                )

            if binding_protocol != server_protocol:
                raise InvalidServerBindingProtocolError(
                    server_name, protocol_str
                )

            server_bindings[binding_protocol] = bindings

        return server_bindings

    return None


def validate_content_type(content_type: str) -> None:
    if content_type != DEFAULT_CONTENT_TYPE:
        raise InvalidContentTypeError(
            content_type, f'valid content types: {DEFAULT_CONTENT_TYPE}'
        )


def validate_asyncapi_version(asyncapi_version: str) -> None:
    if asyncapi_version != ASYNCAPI_VERSION:
        raise InvalidAsyncApiVersionError(
            asyncapi_version, f'valid versions: {ASYNCAPI_VERSION}'
        )


def build_channels(spec: Dict[str, Any]) -> Dict[str, Channel]:
    channels = {}

    for channel_name, channel_spec in spec['channels'].items():
        channels[channel_name] = Channel(
            name=channel_name,
            subscribe=build_operation(channel_spec.pop('subscribe', None)),
            publish=build_operation(channel_spec.pop('publish', None)),
            **channel_spec,
        )

    return channels


def build_operation(
    operation_spec: Optional[Dict[str, Any]]
) -> Optional[Operation]:
    if operation_spec is None:
        return None

    message_spec = operation_spec.get('message')

    return Operation(
        message=build_message(message_spec) if message_spec else None,
        operation_id=operation_spec.get('operationId', None),
    )


def build_message(message_spec: Dict[str, Any]) -> Message:
    content_type = message_spec.get('contentType')

    if content_type:
        validate_content_type(content_type)

    return Message(
        content_type=content_type,
        payload=jsonschema_asdataclass(
            message_spec['name'].replace(' ', ''),
            message_spec.get('payload', {}),
        ),
        **{
            k: v
            for k, v in message_spec.items()
            if k != 'contentType' and k != 'payload'
        },
    )


def build_components(
    components_spec: Optional[Dict[str, Any]]
) -> Optional[Components]:
    if components_spec is None:
        return None

    return Components(
        messages={
            msg_id: build_message(message_spec)
            for msg_id, message_spec in components_spec['messages'].items()
        }
        if 'messages' in components_spec
        else None,
        schemas=components_spec.get('schemas'),
    )


def fill_refs(
    spec: Dict[str, Any], full_spec: Optional[Dict[str, Any]] = None
) -> None:
    if full_spec is None:
        full_spec = spec

    for key, value in spec.items():
        if isinstance(value, dict):
            while '$ref' in value:
                spec[key] = value = dict_from_ref(spec[key]['$ref'], full_spec)

            fill_refs(value, full_spec)


def dict_from_ref(ref: str, spec: Dict[str, Any]) -> Dict[str, Any]:
    ref_keys = deque(ref.strip('#/').split('/'))

    while ref_keys and isinstance(spec, dict):
        spec = spec.get(ref_keys.popleft(), {})

    if not spec:
        raise ReferenceNotFoundError(ref)

    return spec
